/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* This file contains the definitions of types and functions related to the dispatch module
 * and that can be called by extensions. 
 */

/* The dispatch module has two main roles:
 *  - help determine if a message can be handled locally (during the routing process)
 *  - pass the message to the callback(s) that will handle it (during the dispatch process)
 *
 * This module consists in registering callbacks that will be called when a message is received.
 * The decision whether a callback must be called depends on the content of the message and
 * how the callback was registered.
 * 
 * These are the possibilities for registering a callback:
 *
 * -> by AVP value.
 *  This callback will be called when a message is received and contains a certain AVP with a specified value.
 *
 * -> by AVP.
 *  This callback will be called when the received message contains a certain AVP.
 *
 * -> by command-code.
 *  This callback will be called when the message is a specific command.
 *
 * -> by application.
 *  This callback will be called when the message has a specific application-id.
 *
 * ( by vendor: would this be useful? it may be added later)
 *
 * In order to advertise the support for a given application in the CER/CEA messages, 
 * at least one callback must be registered on this application.
 *
 * Note that several criteria may be selected at the same time, for example command-code AND application id.
 *
 * When a callback is called, it receives the message as parameter, and eventually a pointer to 
 * the AVP in the message when this is appropriate.
 *
 * The callback must process the message, and eventually create an answer to it. See the definition
 * bellow for more information.
 *
 * If no callback has handled the message, a default handler will be called with the effect of
 * requeuing the message for forwarding on the network to another peer (for requests), or 
 * discarding the message (for answers).
 */

#ifndef _DISPATCH_API_H
#define _DISPATCH_API_H

#include <waaad/message-api.h>

/* The allowed methods for registering a callback. */
typedef enum {
	DISP_REG_ANY = 1,		/* The callback is called on any message. This should be only used for debug. */
	DISP_REG_APPID,			/* The callback is called for any message with the specified application-id */
	DISP_REG_CC,			/* The callback is called for message of the specified command-code (request or answer). App id may be specified. */
	DISP_REG_AVP,			/* The callback is called for messages containing a specific AVP. Command-code and App id may be specified. */
	DISP_REG_AVP_ENUMVAL		/* The callback is called for messages containing a specific AVP with a specific enumerated value. Command-code and App id may be specified. */
} disp_reg_t;

/* Note on the order of callbacks calls:
  This is the priority of the calls (first in the list = first called).
  If several callbacks are registered with the same disp_reg_t, their order is unspecified.
  
  DISP_REG_ANY
  DISP_REG_AVP_ENUMVAL & DISP_REG_AVP
  DISP_REG_CC
  DISP_REG_APPID
*/

/* Explanation follow */
typedef struct {
	uint32_t	flags; /* 1: registering the Auth part of application; 2: registering the Acct part of application. Only used for DISP_REG_APPID */
	dict_object_t *	app_id;
	dict_object_t *	command;
	dict_object_t *	avp;
	dict_object_t *	value;
} disp_reg_val_t;

/* When a callback is registered, a "when" argument is passed in addition to the disp_reg_t value 
 * to specify which values the criteria must match.
 *
 * Here is the details on the "when" argument, depending on the disp_reg_t value.
 *
 * DISP_REG_ANY.
 *  In this case, "when" must be NULL.
 *
 * In any other case, the "when" parameter points to a valid disp_reg_val_t structure. Detail is given bellow.
 *
 * DISP_REG_APPID.
 *  Only the "app_id" field must be set, other fields are ignored. It points to a dictionary object of type DICT_APPLICATION.
 *
 * DISP_REG_CC.
 *  The "command" field must be defined and point to a dictionary object of type DICT_COMMAND.
 *  The "app_id" may be also set. In the case it is set, it restricts the callback to be called only with this command-code and app id.
 *  The other fields are ignored.
 *
 * DISP_REG_AVP.
 *  The "avp" field of the structure must be set and point to a dictionary object of type DICT_AVP.
 *  The "app_id" field may be set to restrict the messages matching to a specific app id.
 *  The "command" field may also be set to a valid DICT_COMMAND object.
 *  The content of the "value" field is ignored.
 *
 * DISP_REG_AVP_ENUMVAL.
 *  All fields have the same constraints and meaning as in DISP_REG_AVP. In addition, the "value" field must be set
 *  and points to a valid DICT_TYPE_ENUM object. 
 *
 * Here is a sumary of the fields: ( M : must be set; m : may be set, or NULL/0 to ignore; 0 : ignored )
 *
 *  field:     app_id    command     avp    value
 * APPID :       M          0         0       0
 * CC    :       m          M         0       0
 * AVP   :       m          m         M       0
 * AVP_EN:       m          m         M       M
 */
 

/* Here is the prototype of callback functions that will be called */
/*
 * CALLBACK:	disp_cb_t
 *
 * PARAMETERS:
 *  msg     : The location of message that trigged the callback call.
 *  avp     : for callbacks registered with DISP_REG_AVP or DISP_REG_AVP_ENUMVAL, this points to the triggering AVP. NULL otherwise.
 *
 * DESCRIPTION: 
 *   This is the prototype of the callback functions that are registered with dicp_cb_reg function.
 *  See introduction and previous comments for more information on how this callback is called.
 *  
 *  The callback must set *msg to NULL if it creates an answer or "handles" a received answer, so that no further processing is done.
 * 
 * RETURN VALUE:
 *   0	: success. 
 *  !0 	: An error occurred (ex: EINVAL = invalid parameter; ENOMEM: not enough memory, ...).
 *	 In this case, the error is logged in the daemon and the message is discarded.
 */
typedef int (*disp_cb_t) ( msg_t ** msg, msg_avp_t * avp );

/* The following opaque type represents a handler to a registered callback. This allows to remove a registered callback. */
typedef void disp_cb_hdl_t;

#ifndef IN_EXTENSION

/*
 * FUNCTION:	disp_register
 *
 * PARAMETERS:
 *  cb 		  : The callback function to register.
 *  how	  	  : How the callback must be registered.
 *  when          : Values that must match, depending on the how argument.
 *  handle        : On success, a handler to the registered callback is stored here. 
 *		   This handler will be used to unregister the cb.
 *
 * DESCRIPTION: 
 *   Register a new callback to handle messages delivered locally.
 *  When a callback is registered for DISP_REG_APPID, it also have the effect of advertising 
 *  this application support in the CER/CEA messages.
 *
 * RETURN VALUE:
 *  0      	: The callback is registered.
 *  EINVAL 	: A parameter is invalid.
 *  ENOMEM	: Not enough memory to complete the operation
 */
int disp_register ( disp_cb_t cb, disp_reg_t how, disp_reg_val_t * when, disp_cb_hdl_t ** handle );


/*
 * FUNCTION:	disp_unregister
 *
 * PARAMETERS:
 *  handler       : The callback that must be unregistered.
 *
 * DESCRIPTION: 
 *   Removes a callback from the list of registered callbacks.
 *
 * RETURN VALUE:
 *  0      	: The callback is unregistered.
 *  EINVAL 	: A parameter is invalid.
 */
int disp_unregister ( disp_cb_hdl_t * handle );

#endif /* ! IN_EXTENSION */

/******************************************************************/

/* The version of this API, to check binary compatibility -- increment each time a change is made in api_disp_t */
#define WAAAD_API_DISP_VER	1

/* Now define the type of the structure that contains the callback to pass to extensions */
typedef struct {
	/* The header is common to all sub-API pieces */
	size_t	length;		/* The size of this structure, may be useful for extensions not using the facility */
	int	version;	/* The version of this API/ABI, must be WAAAD_API_DISP_VER */
	
	/* the remaining is dispatching-specific */
	int (*disp_register)    ( disp_cb_t cb, disp_reg_t how, disp_reg_val_t * when, disp_cb_hdl_t ** handle );
	int (*disp_unregister)  ( disp_cb_hdl_t * handle );
} api_disp_t;

#ifdef IN_EXTENSION

/* From within the extensions, we register callbacks in the following global structure */
#ifdef DECLARE_API_POINTERS
api_disp_t * g_api_disp=NULL;
#else /* DECLARE_API_POINTERS */
extern api_disp_t * g_api_disp;
#endif /* DECLARE_API_POINTERS */

/* These defines allow to call functions from extension in the same way as in waaad */
#define disp_register		g_api_disp->disp_register
#define disp_unregister		g_api_disp->disp_unregister

#else /* IN_EXTENSION */

/* From the daemon, we must initialize the API object, in extension.c */
# define MY_WAAAD_API_DISP_VER 1
# if MY_WAAAD_API_DISP_VER != WAAAD_API_DISP_VER
#  error "You must update API_INIT_DISP also"
# endif

#define API_INIT_DISP( api_disp ) 				\
{								\
	(api_disp).length           = sizeof(api_disp_t);	\
	(api_disp).version          = WAAAD_API_DISP_VER;	\
	(api_disp).disp_register    = disp_register;		\
	(api_disp).disp_unregister  = disp_unregister;		\
}

#endif /* IN_EXTENSION */

#endif /* _DISPATCH_API_H */


